/*
 *  Copyright (C) 2007 Ludovic Jacomme (ludovic.jacomme@gmail.com)
 *
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Based on the source code of PelDet written by Danzel A.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <SDL.h>

#include "gp2x_psp.h"
#include "psp_global.h"
#include "psp_sdl.h"
#include "psp_danzeff.h"
#include "psp_menu.h"
#include "psp_fmgr.h"
#include "psp_editor.h"


# define EDITOR_MAX_LINES     60000
# define EDITOR_MAX_WIDTH      1024

  typedef struct line_t {
    char* text;
    short width; 
    short max_width;
  } line_t;

  typedef struct text_t {
    line_t*   lines[EDITOR_MAX_LINES];

    int       max_width;
    int       max_lines;

    int       num_lines;
    int       top_line;
    int       left_col;
    int       curr_line;
    int       curr_col;

    u32       last_blink;
    char      cursor_blink;
    char      cursor_moving;
    char      cursor_end;
    int       last_col;
    
  } text_t;

  static text_t* a_text;

  int editor_colors[EDITOR_MAX_COLOR];

  char* editor_colors_name[EDITOR_MAX_COLOR] = {
     "white" ,
     "black" ,
     "dark blue" ,
     "green" ,
     "red" ,
     "brown" ,
     "magenta" ,
     "orange" ,
     "yellow" ,
     "bright green" ,
     "cyan" ,
     "bright blue" ,
     "blue" ,
     "pink" ,
     "gray" ,
     "bright gray",
     "image"
  };


line_t*
psp_editor_add_line(int line_id, char *buffer)
{
  if (line_id >= a_text->max_lines) return 0;

  line_t* a_line = a_text->lines[line_id];
  int length = strlen( buffer );

  if (! a_line) {
    a_line = (line_t*)malloc( sizeof(line_t) );
    memset(a_line, 0, sizeof( line_t ));
    a_text->lines[line_id] = a_line;
  }

  if (length < a_line->max_width) {
    strcpy(a_line->text, buffer);
    a_line->width = length;
  } else {
    if (a_line->text) free(a_line->text);
    a_line->width     = length;
    a_line->max_width = length;
    a_line->text = strdup( buffer );
  }

  if (line_id >= a_text->num_lines) {
    a_text->num_lines = line_id + 1;
  }
  PSPWRITE.is_modified = 1;
  return a_line;
}

void
psp_editor_resize_line(line_t* a_line, int new_width)
{
  PSPWRITE.is_modified = 1;
  if (new_width > a_text->max_width) new_width = a_text->max_width;
  if (new_width > a_line->max_width) {
    a_line->text = realloc(a_line->text, new_width + 1);
    a_line->max_width = new_width;
  }
}

void
psp_editor_reset_text()
{
  int line_id;
  for (line_id = 0; line_id < a_text->num_lines; line_id++) {
    line_t* a_line = a_text->lines[line_id];
    if (a_line) {
      if (a_line->text) {
        free(a_line->text);
      }
      memset(a_line, 0, sizeof(line_t));
    }
  }
  a_text->num_lines = 0;
  a_text->top_line = 0;
  a_text->left_col = 0;
  a_text->curr_line = 0;
  a_text->curr_col = 0;
  a_text->cursor_moving = 0;
  a_text->cursor_blink = 0;
  a_text->cursor_end = 0;
  a_text->last_col = 0;

  psp_editor_add_line(0, "");
}

void
psp_editor_update_column(int update_last)
{
  if (a_text->curr_col >= a_text->max_width) {
    a_text->curr_col = a_text->max_width - 1;
  }
  if ((a_text->curr_col - a_text->left_col) >= PSPWRITE.screen_w) {
    a_text->left_col = a_text->curr_col - PSPWRITE.screen_w + 1;
  }
  if (a_text->curr_col < 0) {
    a_text->curr_col = 0;
  }
  if (a_text->curr_col < a_text->left_col) {
    a_text->left_col = a_text->curr_col;
  }

  line_t* a_line = a_text->lines[a_text->curr_line];
  if (a_line && a_line->width) {
    if (a_line->width != a_text->curr_col) a_text->cursor_end = 0;
    if (! a_text->cursor_end) {
      if (update_last) {
        a_text->last_col = a_text->curr_col;
      }
    }
  }
}

void
psp_editor_update_line()
{
  if (a_text->curr_line >= a_text->num_lines) {
    a_text->curr_line = a_text->num_lines - 1;
  }
  if ((a_text->curr_line - a_text->top_line) >= PSPWRITE.screen_h) {
    a_text->top_line = a_text->curr_line - PSPWRITE.screen_h + 1;
  }
  if (a_text->curr_line < 0) {
    a_text->curr_line = 0;
  }
  if (a_text->curr_line < a_text->top_line) {
    a_text->top_line = a_text->curr_line;
  }
}


void
psp_editor_goto_line_down()
{
  a_text->cursor_moving = 1;

  if (PSPWRITE.move_on_text) {

    if ((a_text->curr_line + 1) < a_text->num_lines) {
      a_text->curr_line++;
      line_t* a_line_next = a_text->lines[a_text->curr_line];
      if (! a_line_next) {
        a_text->curr_col = 0;
      } else
      if (a_text->cursor_end || (a_text->curr_col > a_line_next->width)) {
        a_text->curr_col = a_line_next->width;
        if ((a_text->last_col) && (a_text->curr_col > a_text->last_col)) {
          a_text->curr_col = a_text->last_col;
        }
      } else {
        if (a_text->last_col <= a_line_next->width) {
          a_text->curr_col = a_text->last_col;
        } else {
          a_text->curr_col = a_line_next->width;
        }
      }
    }

  } else {
    a_text->curr_line++;
  }
  psp_editor_update_column(0);
  psp_editor_update_line();
}

void
psp_editor_goto_line_up()
{
  a_text->cursor_moving = 1;

  if (PSPWRITE.move_on_text) {

    if (a_text->curr_line > 0) {
      a_text->curr_line--;
      line_t* a_line_prev = a_text->lines[a_text->curr_line];
      if (! a_line_prev) {
        a_text->curr_col = 0;
      } else
      if (a_text->cursor_end || (a_text->curr_col > a_line_prev->width)) {
        a_text->curr_col = a_line_prev->width;
        if ((a_text->last_col) && (a_text->curr_col > a_text->last_col)) {
          a_text->curr_col = a_text->last_col;
        }
      } else {
        if (a_text->last_col <= a_line_prev->width) {
          a_text->curr_col = a_text->last_col;
        }
      }
    }
  } else {
    a_text->curr_line--;
  }
  psp_editor_update_column(0);
  psp_editor_update_line();
}

void
psp_editor_goto_col_left()
{
  a_text->cursor_moving = 1;

  if (PSPWRITE.move_on_text) {

    line_t* a_line = a_text->lines[a_text->curr_line];
    if (! a_line) {
      a_text->curr_col = 0;
    } else {
      if ((a_text->curr_col == 0) && (a_text->curr_line > 0)) {
        a_text->curr_line--;
        line_t* a_line_prev = a_text->lines[a_text->curr_line];
        if (a_line_prev) {
          a_text->curr_col = a_line_prev->width;
          a_text->last_col = a_text->max_width - 1;
        } else {
          a_text->curr_col = 0;
        }
      } else {
        a_text->curr_col--;
      }
    }

  } else {
    a_text->curr_col--;
  }
  psp_editor_update_line();
  psp_editor_update_column(1);
}

void
psp_editor_goto_col_right()
{
  a_text->cursor_moving = 1;

  if (PSPWRITE.move_on_text) {

    line_t* a_line = a_text->lines[a_text->curr_line];
    if (! a_line) {
      if ((a_text->curr_line + 1) < a_text->num_lines) {
        a_text->curr_line++;
        a_text->curr_col = 0;
      }
    } else {
      if ( (a_text->curr_col == a_line->width          ) && 
           ((a_text->curr_line + 1) < a_text->num_lines) ) {
        a_text->curr_line++;
        a_text->curr_col = 0;
      } else {
        a_text->curr_col++;
      }
    }

  } else {
    a_text->curr_col++;
  }
  psp_editor_update_line();
  psp_editor_update_column(1);
}

void
psp_editor_goto_col_begin()
{
  a_text->cursor_moving = 1;
  a_text->curr_col = 0;
  psp_editor_update_column(1);
}

void
psp_editor_goto_col_end()
{
  a_text->cursor_moving = 1;
  line_t* a_line = a_text->lines[a_text->curr_line];
  if (a_line) {
    a_text->curr_col = a_line->width;
    a_text->last_col = a_text->max_width - 1;
    a_text->cursor_end = 1;
  } else {
    a_text->curr_col = 0;
  }
  psp_editor_update_column(1);
}

void
psp_editor_goto_first_line()
{
  a_text->cursor_moving = 1;
  a_text->curr_col = 0;
  a_text->curr_line = 0;
  psp_editor_update_column(0);
  psp_editor_update_line();
}

void
psp_editor_goto_last_line()
{
  a_text->cursor_moving = 1;
  a_text->curr_col = 0;
  a_text->curr_line = a_text->num_lines;
  psp_editor_update_column(0);
  psp_editor_update_line();
}

void
psp_editor_goto_page_down()
{
  a_text->cursor_moving = 1;
  a_text->curr_line += PSPWRITE.screen_h;
  psp_editor_update_line();
}

void
psp_editor_goto_page_up()
{
  a_text->cursor_moving = 1;
  a_text->curr_line -= PSPWRITE.screen_h;
  psp_editor_update_line();
}

void
psp_editor_goto_word_left()
{
  a_text->cursor_moving = 1;
  line_t* a_line = a_text->lines[a_text->curr_line];
  if (a_line && a_line->text) {
    int index = a_text->curr_col;
    if (index > 0) index--;
    while ((index > 0) && (a_line->text[index] == ' ')) index--;
    while ((index > 0) && (a_line->text[index] != ' ')) index--;
    if (a_line->text[index] == ' ') index++;
    a_text->curr_col = index;
    
  } else a_text->curr_col = 0;
  psp_editor_update_column(1);
}

void
psp_editor_goto_word_right()
{
  a_text->cursor_moving = 1;
  line_t* a_line = a_text->lines[a_text->curr_line];
  if (a_line && a_line->text) {
    int index = a_text->curr_col;
    while ((index < a_line->width) && (a_line->text[index] != ' ')) index++;
    while ((index < a_line->width) && (a_line->text[index] == ' ')) index++;
    a_text->curr_col = index;
    
  } else a_text->curr_col = 0;
  psp_editor_update_column(1);
}

void
psp_editor_insert_char(int line_id, int col_id, int c)
{
  PSPWRITE.is_modified = 1;

  uchar c1 = (c >> 8) & 0xff;
  uchar c2 = c & 0xff;
  if ((c1 == 0xc2) || (c1 == 0xc3)) {
    uchar new_c = psp_convert_utf8_to_iso_8859_1(c1, c2);
    if (new_c) { c = new_c; }
  }

  line_t* a_line = a_text->lines[line_id];
  if (! a_line) {
    a_line = psp_editor_add_line(line_id, "");
  }
  if (col_id >= a_line->width) {
    psp_editor_resize_line(a_line, 1 + ((col_id * 3) / 2) );
    while (a_line->width < col_id) {
      a_line->text[a_line->width++] = ' ';
    }
    a_line->text[col_id] = c;
    a_line->width++;
  } else {
    if ((a_line->width + 1) >= a_line->max_width) {
      psp_editor_resize_line(a_line, 1 + ((a_line->width * 3) / 2) );
    }
    a_line->width++;
    int index = a_line->width;
    while (index > col_id) {
      a_line->text[index] = a_line->text[index - 1];
      index--;
    }
    a_line->text[col_id] = c;
  }
}

void
psp_editor_insert_curr_char(int c)
{
  if ((c == '\t') && PSPWRITE.expand_tab) {
    int tab;
    for (tab = 0; tab < PSPWRITE.tab_stop; tab++) {
      psp_editor_insert_curr_char(' ');
    }
  } else {
    psp_editor_insert_char(a_text->curr_line, a_text->curr_col, c);
    a_text->curr_col++;
  }
  psp_editor_update_column(1);
}

void
psp_editor_split_line(int line_id, int col_id)
{
  if (a_text->num_lines >= a_text->max_lines) return;

  PSPWRITE.is_modified = 1;

  line_t* a_line = a_text->lines[line_id];
  if (! a_line) {
    a_line = psp_editor_add_line(line_id, "");
  }
  int n_line_id = line_id + 1;
  int delta = a_text->num_lines - line_id - 1;
  if (delta > 0) {
    int index = 0;
    for (index = a_text->num_lines - 1; index > line_id; index--) {
      a_text->lines[index + 1] = a_text->lines[index];
    }
    a_text->lines[n_line_id] = 0;
  }
  line_t* n_line = psp_editor_add_line(n_line_id, "");

  if (col_id >= a_line->width) {
    /* nothing more to do */
  } else {
    int delta = a_line->width - col_id;
    psp_editor_resize_line(n_line, 1 + ((delta * 3) / 2));
    int index = col_id;
    while (index < a_line->width) {
      n_line->text[n_line->width++] = a_line->text[index];
      index++;
    }
    a_line->width = col_id;
  }
  a_text->num_lines++;
}

void
psp_editor_split_curr_line()
{
  psp_editor_split_line(a_text->curr_line, a_text->curr_col);
  a_text->curr_col = 0;
  a_text->curr_line++;
  psp_editor_update_column(1);
  psp_editor_update_line();
}

void
psp_editor_delete_char(int col_id)
{
  PSPWRITE.is_modified = 1;

  line_t* a_line = a_text->lines[a_text->curr_line];
  if (! a_line) {
    a_line = psp_editor_add_line(a_text->curr_line, "");
  }
  if (col_id > a_line->width) {
    /* Nothing to do */
  } else 
  if (col_id > 0) {
    int delta = a_line->width - col_id;
    if (delta > 0) {
      char* scan_text = &a_line->text[col_id];
      memcpy( scan_text - 1, scan_text, delta);
    }
    a_line->width--;
    a_line->text[a_line->width] = 0;
  }
}

void
psp_editor_suppr_char(int col_id)
{
  PSPWRITE.is_modified = 1;

  line_t* a_line = a_text->lines[a_text->curr_line];
  if (! a_line) {
    a_line = psp_editor_add_line(a_text->curr_line, "");
  }
  if (col_id > a_line->width) {
    /* Nothing to do */
  } else {
    int delta = a_line->width - col_id;
    if (delta > 0) {
      char* scan_text = &a_line->text[col_id];
      memcpy( scan_text, scan_text + 1, delta);
    }
    a_line->width--;
    a_line->text[a_line->width] = 0;
  }
}

void
psp_editor_delete_line(int line_id)
{
  PSPWRITE.is_modified = 1;

  line_t* a_line = a_text->lines[line_id];
  if (a_line) {
    if (a_line->text) free(a_line->text);
    a_line->text = 0;
    free(a_line);
    a_text->lines[line_id] = 0;
    a_line = 0;
  }
  int delta = a_text->num_lines - line_id - 1;
  if (delta > 0) {
    memcpy(&a_text->lines[line_id], &a_text->lines[line_id +1], sizeof(line_t *) * delta);
    a_text->lines[a_text->num_lines - 1] = 0;
  }
  a_text->num_lines--;
  if (a_text->num_lines <= 0) {
    psp_editor_add_line(0, "");
  }
}

void
psp_editor_join_prev_line(int a_line_id)
{
  /* Join with previous line */

  int p_line_id = a_line_id - 1;
  line_t* a_line = a_text->lines[a_line_id];
  line_t* p_line = a_text->lines[p_line_id];
  if (! a_line) {
    a_line = psp_editor_add_line(a_line_id, "");
  }
  if (! p_line) {
    p_line = psp_editor_add_line(p_line_id, "");
  }
  int a_width = a_line->width;
  int p_width = p_line->width;
  int r_width = p_width + a_width;
  psp_editor_resize_line(p_line, r_width);
  if (r_width > p_line->max_width) r_width = p_line->max_width;
  int delta = r_width - p_width;
  if (delta > 0) {
    memcpy(&p_line->text[p_width], a_line->text, delta);
  }
  p_line->width = r_width;
  a_text->curr_col  = p_width;
  a_text->curr_line = p_line_id;
  psp_editor_delete_line(a_line_id);
}

void
psp_editor_delete_curr_char()
{
  if (a_text->curr_col > 0) {
    psp_editor_delete_char(a_text->curr_col);
    a_text->curr_col--;
  } else 
  if (a_text->curr_line > 0) {

    psp_editor_join_prev_line(a_text->curr_line);

    psp_editor_update_column(1);
    psp_editor_update_line();
  }
}

void
psp_editor_suppr_curr_char()
{
  line_t* a_line = a_text->lines[a_text->curr_line];
  if (! a_line) {
    a_line = psp_editor_add_line(a_text->curr_line, "");
  }
  if (a_text->curr_col >= a_line->width) {
    psp_editor_join_prev_line(a_text->curr_line + 1);
  } else {
    psp_editor_suppr_char(a_text->curr_col);
  }
}

void
psp_editor_clear_curr_line()
{
  int line_id = a_text->curr_line;
  line_t* a_line = a_text->lines[line_id];
  if (! a_line) {
    a_line = psp_editor_add_line(line_id, "");
  }
  a_line->width = 0;
  a_line->text[0] = 0;
  a_text->curr_col = 0;
  psp_editor_update_column(1);
}

void
psp_editor_new()
{
  psp_editor_reset_text();
  PSPWRITE.is_modified = 0;
  PSPWRITE.ask_overwrite = 1;

  strcpy(PSPWRITE.edit_filename, PSPWRITE.psp_txtdir);
  strcat(PSPWRITE.edit_filename, "newfile.txt");
}

int
psp_editor_load(char *filename)
{
  char  Buffer[EDITOR_MAX_WIDTH];
  char  Line[EDITOR_MAX_WIDTH + 32];
  FILE* FileDesc;

  psp_editor_reset_text();
  PSPWRITE.is_modified = 0;

  FileDesc = fopen(filename, "r");
  if (FileDesc != (FILE *)0) {

    strncpy(PSPWRITE.edit_filename, filename, MAX_PATH);

    int line_id = 0;
    while (fgets(Buffer, a_text->max_width, FileDesc) != (char *)0) {
      char *Scan = strchr(Buffer,'\n');
      if (Scan) *Scan = '\0';
      /* For this #@$% of windows ! */
      Scan = strchr(Buffer,'\r');
      if (Scan) *Scan = '\0';

      /* Copy the buffer into line */ 
      int tab;
      int index;
      int target = 0;
      for (index = 0; Buffer[index]; index++) {
        uchar c = (uchar)Buffer[index];
        /* expand tabs */
        if (c == '\t') {
          if (PSPWRITE.expand_tab) {
            for (tab = 0; tab < PSPWRITE.tab_stop; tab++) {
              Line[target++] = ' ';
            }
          } else Line[target++] = c;

        } else {

          if (c < ' ') c = ' ';
          Line[target++] = c;
        }
        if (target >= EDITOR_MAX_WIDTH) break;
      }
      Line[target] = 0;

      psp_editor_add_line(line_id, Line);
      line_id++;
      if (line_id >= a_text->max_lines) break;
    }
    PSPWRITE.is_modified = 0;
    fclose(FileDesc);

    return 0;
  }
  return 1;
}

int
psp_editor_save()
{
  char  Buffer[EDITOR_MAX_WIDTH];
  FILE* FileDesc;

  PSPWRITE.is_modified = 0;

  /* check if extention is present */
  if (PSPWRITE.dos_mode) {
    char* filename = PSPWRITE.edit_filename;
    int length = strlen(filename);
    char* last_dot = strrchr(filename, '.');
    if (! last_dot || ((last_dot - filename) < (length - 4))  ) {
      if (length < (MAX_PATH-4)) {
        strcat(filename, ".txt");
      }
    }
  }
  /* check if fullpath ? */
  {
    char* found_slash = strchr(PSPWRITE.edit_filename, '/');
    if (! found_slash) {
      char TmpFileName[MAX_PATH];
      strcpy(TmpFileName, PSPWRITE.psp_txtdir);
      strcat(TmpFileName, PSPWRITE.edit_filename);
      strcpy(PSPWRITE.edit_filename, TmpFileName);
    }
  }

  FileDesc = fopen(PSPWRITE.edit_filename, "w");

  if (FileDesc != (FILE *)0) {

    int line_id = 0;
    int last_line_empty = 1;
    for (line_id = 0; line_id < a_text->num_lines; line_id++) {
      line_t* a_line = a_text->lines[line_id];
      if (a_line && a_line->width) { 
        memcpy(Buffer, a_line->text, a_line->width);
        Buffer[a_line->width] = 0;
        fputs(Buffer, FileDesc);
        last_line_empty = 0;
      } else last_line_empty = 1;
      if (PSPWRITE.dos_mode) {
        fprintf(FileDesc, "\r");
      }
      fprintf(FileDesc, "\n");
    }
    if (last_line_empty) {
      if (PSPWRITE.dos_mode) {
        fprintf(FileDesc, "\r");
      }
      fprintf(FileDesc, "\n");
    }
    fclose(FileDesc);
    return 0;
  }
  return 1;
}

void 
psp_editor_init()
{
  if (! a_text) {
    a_text = (text_t *)malloc(sizeof( text_t ));
  }
  memset(a_text, 0, sizeof( text_t ));

  a_text->max_width = EDITOR_MAX_WIDTH;
  a_text->max_lines = EDITOR_MAX_LINES;

  psp_editor_new();

  /* initialize color */
  editor_colors[COLOR_WHITE         ] = psp_sdl_rgb(255, 255, 255);
  editor_colors[COLOR_BLACK         ] = psp_sdl_rgb(0, 0, 0);
  editor_colors[COLOR_DARK_BLUE     ] = psp_sdl_rgb(0, 0, 0x55);
  editor_colors[COLOR_GREEN         ] = psp_sdl_rgb(0, 0xAA, 0);
  editor_colors[COLOR_RED           ] = psp_sdl_rgb(0xAA, 0, 0);
  editor_colors[COLOR_BROWN         ] = psp_sdl_rgb(0xAA, 0x55, 0);
  editor_colors[COLOR_MAGENTA       ] = psp_sdl_rgb(0xAA, 0, 0xAA);
  editor_colors[COLOR_ORANGE        ] = psp_sdl_rgb(255, 0xAA, 0);
  editor_colors[COLOR_YELLOW        ] = psp_sdl_rgb(255, 255, 0x55);
  editor_colors[COLOR_BRIGHT_GREEN  ] = psp_sdl_rgb(0, 255, 0);
  editor_colors[COLOR_CYAN          ] = psp_sdl_rgb(0, 0xAA, 0xAA);
  editor_colors[COLOR_BRIGHT_BLUE   ] = psp_sdl_rgb(0x55, 0x55, 255);
  editor_colors[COLOR_BLUE          ] = psp_sdl_rgb(0, 0, 0xAA);
  editor_colors[COLOR_PINK          ] = psp_sdl_rgb(255, 0, 255);
  editor_colors[COLOR_GRAY          ] = psp_sdl_rgb(0x55, 0x55, 0x55);
  editor_colors[COLOR_BRIGHT_GRAY   ] = psp_sdl_rgb(0xAA, 0xAA, 0xAA);
  editor_colors[COLOR_IMAGE         ] = psp_sdl_rgb(0xff, 0xff, 0xff);

  psp_sdl_select_font_6x10();
}

# define GP2X_EDITOR_BLINK_TIME 500

void
psp_display_cursor()
{
  psp_sdl_select_font_6x10();

  int real_y_min = (GP2X_SDL_SCREEN_HEIGHT - (PSPWRITE.screen_h * psp_font_height)) / 2;
  int real_x_min = (GP2X_SDL_SCREEN_WIDTH  - (PSPWRITE.screen_w * psp_font_width)) / 2;

  /* Display cursor */
  int fg_color = editor_colors[PSPWRITE.fg_color];
  u32 curr_clock = SDL_GetTicks();
  if ((curr_clock - a_text->last_blink) > GP2X_EDITOR_BLINK_TIME) {
    a_text->last_blink = curr_clock;
    a_text->cursor_blink = ! a_text->cursor_blink;
  }
  if (a_text->cursor_blink || a_text->cursor_moving) {
    int c_y = a_text->curr_line - a_text->top_line;
    int c_x = a_text->curr_col - a_text->left_col;
    if ((c_y >= 0) && (c_x >= 0)) {
      int c_real_y = real_y_min + c_y * psp_font_height;
      int c_real_x = real_x_min + c_x * psp_font_width;
      psp_sdl_fill_rectangle(c_real_x, c_real_y, 0, psp_font_height, fg_color, 0);
    }
  }
}

void
psp_display_editor()
{
  psp_sdl_select_font_6x10();

  int line_from = a_text->top_line;
  int line_to   = line_from + PSPWRITE.screen_h;
  if (line_to > a_text->num_lines) line_to = a_text->num_lines;

  int line_id;

  int real_y_min = (GP2X_SDL_SCREEN_HEIGHT - (PSPWRITE.screen_h * psp_font_height)) / 2;
  int real_x_min = (GP2X_SDL_SCREEN_WIDTH  - (PSPWRITE.screen_w * psp_font_width)) / 2;

  int y = 0;
  int real_y = real_y_min;
  int fg_color = editor_colors[PSPWRITE.fg_color];

  if (PSPWRITE.bg_color == COLOR_IMAGE) {
    psp_sdl_blit_editor();
  } else {
    int bg_color = editor_colors[PSPWRITE.bg_color];
    psp_sdl_clear_screen(bg_color);
  }

  for (line_id = line_from; line_id < line_to; line_id++) {
    line_t* a_line = a_text->lines[line_id];
    if (a_line && a_line->text) {
      int col_from = a_text->left_col;
      int col_to   = a_line->width;
      int col_id;
      int x = 0;
      int real_x = real_x_min;
      for (col_id = col_from; col_id < col_to; col_id++) {
        unsigned char c = a_line->text[col_id];
        if (c == '\t') c = 1;
        else if (c < ' ') c = ' ';
        psp_sdl_put_char(real_x, real_y, fg_color, 0, c, 1, 0);
        real_x += psp_font_width;
        x++;
        if (x >= PSPWRITE.screen_w) break;
      }
    }
    real_y += psp_font_height;
    y++;
    if (y > PSPWRITE.screen_h) break;
  }
}

# define GP2X_EDITOR_MIN_MOVE_TIME   10000
# define GP2X_EDITOR_REPEAT_TIME    300000
# define GP2X_EDITOR_MIN_EDIT_TIME  180000
# define GP2X_EDITOR_MIN_TIME       150000
void
psp_editor_main_loop()
{
  int  danzeff_mode;
  int  danzeff_key;
  long old_pad;
  long new_pad;
  int  last_time;
  int  repeat_mode;

  gp2xCtrlData c;

  old_pad      = 0;
  last_time    = 0;
  danzeff_mode = 0;
  repeat_mode  = 0;

  psp_editor_init();

  psp_edit_main_menu();

  while (1) {

    psp_display_editor();
    psp_display_cursor();

    if (danzeff_mode) {
      danzeff_moveTo(0, 0);
      danzeff_render();
    }
    psp_sdl_flip();

    while (1) {

      gp2xCtrlPeekBufferPositive(&c, 1);
      c.Buttons &= GP2X_ALL_BUTTON_MASK;

      new_pad = c.Buttons;

      if (old_pad != new_pad) {
        repeat_mode = 0;
        last_time = c.TimeStamp;
        old_pad = new_pad;
        break;

      } else 
      if (new_pad != 0) {
        if ((c.TimeStamp - last_time) > GP2X_EDITOR_REPEAT_TIME) {
          repeat_mode = 1;
        }
      } else {
        repeat_mode = 0;
      }

      if (repeat_mode) {
        if ((c.TimeStamp - last_time) > GP2X_EDITOR_MIN_MOVE_TIME) {
          last_time = c.TimeStamp;
          old_pad = new_pad;
          break;
        }
      } else {

        if ((c.TimeStamp - last_time) > GP2X_EDITOR_MIN_TIME) {
          last_time = c.TimeStamp;
          old_pad = new_pad;
          break;
        }
      }
    }

    if (danzeff_mode) {

      danzeff_key = danzeff_readInput(c);

      if (danzeff_key > DANZEFF_START) {

        if (danzeff_key >= ' ') {
          psp_editor_insert_curr_char(danzeff_key);
        } else
        if (danzeff_key == DANZEFF_DEL) {
          psp_editor_delete_curr_char();
        } else
        if (danzeff_key == DANZEFF_TAB) {
          psp_editor_insert_curr_char('\t');
        } else
        if (danzeff_key == DANZEFF_CLEAR) {
          psp_editor_clear_curr_line();
        } else
        if (danzeff_key == DANZEFF_PAGE_UP) {
          psp_editor_goto_page_up();
        } else
        if (danzeff_key == DANZEFF_PAGE_DOWN) {
          psp_editor_goto_page_down();
        } else
        if (danzeff_key == DANZEFF_END) {
          psp_editor_goto_col_end();
        } else
        if (danzeff_key == DANZEFF_HOME) {
          psp_editor_goto_first_line();
        } else
        if (danzeff_key == DANZEFF_ENTER) {
          psp_editor_split_curr_line();
        }
# if 0
        else
        if (danzeff_key == DANZEFF_UP) {
          psp_editor_goto_col_begin();
        } else
        if (danzeff_key == DANZEFF_DOWN) {
          psp_editor_split_curr_line();
        } else
        if (danzeff_key == DANZEFF_LEFT) {
          psp_editor_goto_col_left();
        } else
        if (danzeff_key == DANZEFF_RIGHT) {
          psp_editor_goto_col_right();
        }
# endif

      } else
      if ((danzeff_key == DANZEFF_START ) || 
          (danzeff_key == DANZEFF_SELECT)) 
      {
        danzeff_mode = 0;
        old_pad = new_pad = 0;

        psp_kbd_wait_no_button();
      }

      continue;
    }

    if (!c.Buttons) {
      a_text->cursor_moving = 0;
    }

    if ((c.Buttons & (GP2X_CTRL_LTRIGGER|GP2X_CTRL_RTRIGGER|GP2X_CTRL_START)) ==
        (GP2X_CTRL_LTRIGGER|GP2X_CTRL_RTRIGGER|GP2X_CTRL_START)) {
      /* Exit ! */
      psp_sdl_exit(0);
    } else
    if(new_pad & GP2X_CTRL_START) {
      danzeff_mode = 1;
    } else
    if(new_pad & GP2X_CTRL_SELECT) {
      psp_edit_main_menu();
    } else
    if (new_pad & GP2X_CTRL_RTRIGGER) {

      if(new_pad & GP2X_CTRL_UP) {
        psp_editor_goto_page_up();
      } else
      if(new_pad & GP2X_CTRL_DOWN) {
        psp_editor_goto_page_down();
      } else
      if(new_pad & GP2X_CTRL_RIGHT) {
        psp_editor_goto_word_right();
      } else
      if(new_pad & GP2X_CTRL_LEFT) {
        psp_editor_goto_word_left();
      } else

      if(new_pad & GP2X_CTRL_TRIANGLE) {
         //???
      } else
      if(new_pad & GP2X_CTRL_CROSS) {
         //???
      } else
      if(new_pad & GP2X_CTRL_CIRCLE) {
         //???
      } else
      if(new_pad & GP2X_CTRL_SQUARE) {
         //???
      }

    } else
    if (new_pad & GP2X_CTRL_LTRIGGER) {

      if(new_pad & GP2X_CTRL_UP) {
        psp_editor_goto_first_line();
      } else
      if(new_pad & GP2X_CTRL_DOWN) {
        psp_editor_goto_last_line();
      } else
      if(new_pad & GP2X_CTRL_RIGHT) {
        psp_editor_goto_col_end();
      } else
      if(new_pad & GP2X_CTRL_LEFT) {
        psp_editor_goto_col_begin();
      } else

      if(new_pad & GP2X_CTRL_TRIANGLE) {
        psp_editor_goto_first_line();
      } else
      if(new_pad & GP2X_CTRL_CROSS) {
        psp_editor_goto_last_line();
      } else
      if(new_pad & GP2X_CTRL_CIRCLE) {
        psp_editor_goto_col_end();
      } else
      if(new_pad & GP2X_CTRL_SQUARE) {
        psp_editor_goto_col_begin();
      }

    } else
    if(new_pad & GP2X_CTRL_UP) {
      psp_editor_goto_line_up();
    } else
    if(new_pad & GP2X_CTRL_DOWN) {
      psp_editor_goto_line_down();
    } else
    if(new_pad & GP2X_CTRL_RIGHT) {
      psp_editor_goto_col_right();
    } else
    if(new_pad & GP2X_CTRL_LEFT) {
      psp_editor_goto_col_left();
    } else
    if(new_pad & GP2X_CTRL_TRIANGLE) {
      psp_editor_delete_curr_char();
    } else
    if(new_pad & GP2X_CTRL_SQUARE) {
      psp_editor_suppr_curr_char();
    } else
    if(new_pad & GP2X_CTRL_CIRCLE) {
      psp_editor_insert_curr_char(' ');
    } else
    if(new_pad & GP2X_CTRL_CROSS) {
      psp_editor_split_curr_line();
    }
  }
}
